package simulator;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.*;

public class SimplePeer extends Thread implements Peer{
	
	public static final int MAX_N = 5;
	public static final int TTLPing = 3;
	public static final int TTLQuery = 6;
	
	protected Network net;
	protected int address;
	private LinkedBlockingQueue<Message> queue;
	
	/* status */
	private ArrayList<Integer> knownPeers; //neighbours
	
	
	public SimplePeer(Network net){
		this.net = net;
		this.address = net.addHost(this);
		queue = new LinkedBlockingQueue<Message>();
		knownPeers = new ArrayList<Integer>();
		this.start();
	}
	
	public void joinP2P() {
	    int addressKnownPeer;
	    do {
	        addressKnownPeer = net.getRandomPeerAddress();
	    } while (addressKnownPeer == this.address);
	    this.joinP2P(addressKnownPeer);
	}
	
	public void joinP2P(int addressKnownHost) {
		Message ping = new Message(addressKnownHost);
		
		ping.type = Message.PING;
		ping.field2 = this.address;
		ping.field3 = SimplePeer.TTLPing; //TTL
		
		net.sendMessage(ping);
	}

	
	public int getAddress() {
		return address;
	}
	
	public String lookup(int key) {
	    return "";
	}
	
	public void receiveMessage(Message msg){
		queue.add(msg);
	}
	
	public void run() {
		while(true) {
		    Message msg=null;
		    try {
			    msg = queue.take();
		    } catch (InterruptedException e) {e.printStackTrace();};
		    switch (msg.type) {
		        case Message.PING: this.receivedPing(msg); break;
		        case Message.PONG: this.receivedPong(msg); break;
		    }
		}
	}
	
	protected void receivedPing(Message msg) {
		
		//Forward the ping message to neightbours
        if (msg.field3 > 1) {
		    Iterator<Integer> iter = this.knownPeers.iterator();
		    Message fwdPing;
		    while (iter.hasNext()) {
			    fwdPing = new Message(iter.next());
			    fwdPing.type = Message.PING;
			    fwdPing.field3 = msg.field3 - 1;
			    fwdPing.field2 = msg.field2;
			    if (fwdPing.field2 != fwdPing.getDestination()) {
			        this.net.sendMessage(fwdPing);
			    }
		    }
        }
        
        //with an empty slot 
        if (this.knownPeers.size() < MAX_N && !this.knownPeers.contains(msg.field2)) {
        	//add the new peer to the list of known peers
        	this.knownPeers.add(msg.field2);
        }
        
        //Send a PONG back
        
        Message pong = new Message(msg.field2); //receiver is the PING sender
        pong.field2 = this.address;
        pong.type = Message.PONG;
        pong.field3 = 1;
        this.net.sendMessage(pong);	
	}
		

    protected void receivedPong(Message msg) {
        /*With empty slots and new peers add it as neighbour*/
    	if (this.knownPeers.size() < MAX_N && 
    			!this.knownPeers.contains(msg.field2) &&
    			msg.field2 != this.address) {
    		this.knownPeers.add(msg.field2);
    		
    	}
    }
    
    public void printNeighbours() {
    	String s = "Neighbours of peer " + this.address + ":";
    	
	    Iterator<Integer> iter = this.knownPeers.iterator();
	    while (iter.hasNext()) {
	    	s = s + " " + iter.next();
	    }
	    System.out.println(s);
    }
	
}